# pylint: disable=missing-docstring, invalid-name
__version__ = '0.3.0'

# Provide DepreciationWarning for older Django versions
import warnings

import django

if '1.11' in django.__version__:
    warnings.warn(
        (
            'Django 1.11 LTS and django-helcim will stop receiving '
            'support in April 2020. Ensure you have updated your '
            'versions before then.'
        ),
        DeprecationWarning
    )

# Django configuration details
default_app_config = 'helcim.apps.HelcimConfig'



"""Application configuration file for django-helcim."""
from django.apps import AppConfig


class HelcimConfig(AppConfig):
    """Configuration details for django-helcim."""
    name = 'helcim'
    verbose_name = 'django-helcim'



"""Process and validates data to and from the Helcim API."""

from datetime import datetime
from decimal import Decimal
import logging

LOG = logging.getLogger(__name__)

class Field(object):
    """A single API field.

    Parameters:
        api_name (str): The Helcim API field name for the field.
        field_type (str): The field type: ``s`` (string), ``c``
            (decimal), ``i`` (integer), ``b`` (boolean), ``d`` (date),
            or ``t`` (time).
        min (int, optional): The minimum field length.
        max (int, optional): The maximum field length.
    """

    def __init__(
            self, api_name, field_type='s',
            min_length=None, max_length=None
    ):
        # Ensure a valid field type is provided
        if field_type not in ['s', 'c', 'i', 'b', 'd', 't']:
            raise ValueError(
                'Invalid field type provided for {}: {}'.format(
                    api_name, field_type
                )
            )

        self.field_name = api_name
        self.field_type = field_type
        self.min = min_length
        self.max = max_length

TO_API_FIELDS = {
    'amount': Field('amount', 'c'),
    'amount_shipping': Field('amountShipping', 'c'),
    'amount_tax': Field('amountTax', 'c'),
    'billing_business_name': Field('billing_businessName', 's'),
    'billing_city': Field('billing_city', 's'),
    'billing_contact_name': Field('billing_contactName', 's'),
    'billing_country': Field('billing_country', 's'),
    'billing_email': Field('billing_email', 's'),
    'billing_fax': Field('billing_fax', 's'),
    'billing_phone': Field('billing_phone', 's'),
    'billing_postal_code': Field('billing_postalCode', 's'),
    'billing_province': Field('billing_province', 's'),
    'billing_street_1': Field('billing_street1', 's'),
    'billing_street_2': Field('billing_street2', 's'),
    'cc_address': Field('cardHolderAddress', 's'),
    'cc_cvv': Field('cardCVV', 's', 3, 4),
    'cc_expiry': Field('cardExpiry', 's', 4, 4),
    'cc_name': Field('cardHolderName', 's'),
    'cc_number': Field('cardNumber', 's', 16, 16),
    'cc_postal_code': Field('cardHolderPostalCode', 's'),
    'comments': Field('comments', 's'),
    'customer_code': Field('customerCode', 's'),
    'ecommerce': Field('ecommerce', 'b'),
    'ip_address': Field('ipAddress', 's', 7, 45),
    'mag': Field('cardMag', 's'),
    'mag_enc': Field('cardMagEnc', 's'),
    'mag_enc_serial_number': Field('serialNumber', 's'),
    'product_id': Field('productId', 'i', 1),
    'shipping_business_name': Field('shipping_businessName', 's'),
    'shipping_city': Field('shipping_city', 's'),
    'shipping_contact_name': Field('shipping_contactName', 's'),
    'shipping_country': Field('shipping_country', 's'),
    'shipping_email': Field('shipping_email', 's'),
    'shipping_fax': Field('shipping_fax', 's'),
    'shipping_method': Field('shipping_method', 's'),
    'shipping_phone': Field('shipping_phone', 's'),
    'shipping_postal_code': Field('shipping_postalCode', 's'),
    'shipping_province': Field('shipping_province', 's'),
    'shipping_street_1': Field('shipping_street1', 's'),
    'shipping_street_2': Field('shipping_street2', 's'),
    'tax_details': Field('taxDetails', 's'),
    'test': Field('test', 'b'),
    'token': Field('cardToken', 's', 22, 23),
    'token_f4l4': Field('cardF4L4', 's', 8, 8),
    'token_f4l4_skip': Field('cardF4L4Skip', 'b'),
    'transaction_id': Field('transactionId', 'i', 0),
}

FROM_API_FIELDS = {
    'amount': Field('amount', 'c'),
    'availability': Field('availability', 'b'),
    'approvalCode': Field('approval_code', 's'),
    'avsResponse': Field('avs_response', 's'),
    'cardHolderName': Field('cc_name', 's'),
    'cardNumber': Field('cc_number', 's'),
    'cardToken': Field('token', 's'),
    'cardType': Field('cc_type', 's'),
    'currency': Field('currency', 's'),
    'customerCode': Field('customer_code', 's'),
    'cvvResponse': Field('cvv_response', 's'),
    'date': Field('transaction_date', 'd'),
    'expiryDate': Field('cc_expiry', 's'),
    'orderNumber': Field('order_number', 's'),
    'time': Field('transaction_time', 't'),
    'transactionId': Field('transaction_id', 'i'),
    'type': Field('transaction_type', 's'),
}

def validate_request_fields(details):
    """Validates and coerces request field data prior to submission.

    Uses the TO_API_FIELDS dictionary to determine type and required
    validation to apply.

    Parameters:
        details (dict): The values to validated and coerce.

    Returns:
        dict: The cleaned data (with the same keys as ``details``).

    Raises:
        ValueError: Raised if data fails validation.
    """
    cleaned = {}

    for field_name, field_value in details.items():
        validation = TO_API_FIELDS[field_name]

        # Coerce value and any validation
        # String Types
        if validation.field_type == 's':
            cleaned_value = str(field_value)

            if validation.min and len(cleaned_value) < validation.min:
                raise ValueError(
                    '{} field length too short.'.format(field_name)
                )

            if validation.max and len(cleaned_value) > validation.max:
                raise ValueError(
                    '{} field length too long.'.format(field_name)
                )

        # Integer types
        elif validation.field_type == 'i':
            cleaned_value = int(field_value)

            if validation.min and cleaned_value < validation.min:
                raise ValueError(
                    '{} field value too small.'.format(field_name)
                )

            if validation.max and cleaned_value > validation.max:
                raise ValueError(
                    '{} field value too large.'.format(field_name)
                )

        # Decimal types
        elif validation.field_type == 'c':
            cleaned_value = Decimal(str(field_value))

            if validation.min and cleaned_value < validation.min:
                raise ValueError(
                    '{} field value too small.'.format(field_name)
                )

            if validation.max and cleaned_value > validation.max:
                raise ValueError(
                    '{} field value too large.'.format(field_name)
                )

        # Boolean types
        elif validation.field_type == 'b':
            cleaned_value = 1 if bool(field_value) else 0

        # Add the field to the cleaned data
        cleaned[field_name] = cleaned_value

    return cleaned

def process_request_fields(api, cleaned, additional=None):
    """Converts all data to proper Helcim API request fields.

    Parameters:
        api (dict): The API setting fields.
        cleaned(dict): The cleaned POST data.
        additional: Any additional POST data to directly add (skips
            any conversion steps).

    Returns:
        dict: The data ready for a POST request.
    """
    request_data = {}

    # API Data
    request_data['accountId'] = str(api['account_id'])
    request_data['apiToken'] = str(api['token'])
    request_data['terminalId'] = str(api['terminal_id'])

    # Cleaned Data
    for field_name, field_value in cleaned.items():
        request_data[TO_API_FIELDS[field_name].field_name] = str(field_value)

    # Combine with the additional data
    if additional:
        request_data.update(additional)

    # TODO: Decide what happens with fields not in the TO_API_FIELDS dictionary

    return request_data

def create_raw_request(data):
    """Converts the raw request into a POST string.

    Parameters:
        data (dict): The POST request data dictionary.

    Returns:
        str: The POST request data as a formatted string.
    """
    if data:
        raw_strings = []

        for key, item in data.items():
            raw_strings.append('{}={}'.format(key, item))

        return '&'.join(raw_strings)

    return data

def process_api_response(response, raw_request=None, raw_response=None):
    """Updates API field names/types, and adds additional data.

    Parameters:
        response (str): The API response (as an ``OrderedDict``).
        raw_request (str): Raw request (as string) submitted to API.
        raw_response (str): Raw response (as string) returned by API.

    Returns:
        dict: The validated and convereted API response.
    """
    # Add the standard API response details
    processed = {
        'transaction_success': (
            True if int(response['response']) == 1 else False
        ),
        'response_message': str(response['responseMessage']),
        'notice': str(response['notice']),
    }

    # Add and coerece any fields returned in the transaction field
    if 'transaction' in response:
        for field_name, field_value in response['transaction'].items():

            try:
                api_field = FROM_API_FIELDS[field_name]
                new_name = api_field.field_name

                # String Field
                if api_field.field_type == 's':
                    processed[new_name] = str(field_value)

                # Decimal Field
                elif api_field.field_type == 'c':
                    processed[new_name] = Decimal(field_value)

                # Integer Field
                elif api_field.field_type == 'i':
                    processed[new_name] = int(field_value)

                # Boolean Field
                elif api_field.field_type == 'b':
                    processed[new_name] = True if field_value == '1' else False

                # Date Field
                elif api_field.field_type == 'd':
                    processed[new_name] = datetime.strptime(
                        field_value, '%Y-%m-%d'
                    ).date()

                # Time Field
                elif api_field.field_type == 't':
                    processed[new_name] = datetime.strptime(
                        field_value, '%H:%M:%S'
                    ).time()

                # Handle any invalid types (should never happen...)
                else:
                    LOG.warning(
                        'Field %s has invalid type: %s',
                        api_field.field_name,
                        api_field.field_type,
                    )
                    processed[field_name] = field_value
            except KeyError:
                LOG.warning(
                    'Response field not in FROM_API_FIELDS: %s', field_name
                )
                processed[field_name] = field_value

        # If possible, create the F4L4 field
        cc_number = processed.get('cc_number')

        if cc_number:
            processed['token_f4l4'] = '{}{}'.format(
                cc_number[:4], cc_number[-4:]
            )
        else:
            processed['token_f4l4'] = None

    # Add additional audit information
    processed['raw_request'] = create_raw_request(raw_request)
    processed['raw_response'] = raw_response

    return processed



"""Interface functions with the Helcim Commerce API.

These functions provide an agonstic interface with the Helcim Commerce
API and should work in any application.
"""
from calendar import monthrange
import copy
from datetime import datetime
import re

import pytz
import requests
import xmltodict

from django.conf import settings as django_settings
from django.core import exceptions as django_exceptions
from django.db import IntegrityError

from helcim import conversions, exceptions as helcim_exceptions, models


class BaseRequest():
    """Base class to handle validation and submission to Helcim API.

    Parameters:
        api_details (dict): Details to connect to Helcim API:

            - **url** (*str*): API URL.
            - **account_id** (*str*): Helcim account ID.
            - **token** (*str*): Helcim API token.
            - **terminal_id** (*str*): Helcim terminal ID.

        **kwargs (dict): Any additional transaction details.

    Keyword Arguments:
        amount (dec, optional): The amount for the transaction.
        currency (str, optional): The currency for the transaction.
        cc_name (str, optional): Name of the credit cardholder.
        cc_number (str, optional): 16 digit credit card number.
        cc_expiry (str, optional): 4 digit (MMYY) credit card
            expiry.
        cc_cvv (str, optional): 3 or 4 digit credit card CVV.
        cc_address (str, optional): Address of the credit
            cardholder.
        cc_postal_code (str, optional): Postal code/zip code of
            the credit cardholder.
        customer_code (str, optional): Helcim customer code.
        token (str, optional): 23 digit Helcim card token.
        token_f4l4 (str, optional): 8 digit "first four digits and
            last four digits" of the credit card number
        token_f4l4_skip (bool, optional): Whether to skip the F4L4
            verification.
        mag (string, optional): Non-encrypted credit card magnetic
            strip data.
        mag_enc (str, optional): Encrypted credit card magnestic
            strip data.
        mag_enc_serial_number (string, optional): Terminal serial
            number.
        order_number (str, optional): An assigned order number for
            the purchase.
        ecommerce (bool, optional): Whether this is an e-commerce
            transaction or not.
        comments (str, optional): Any additional comments with this
            transaction.
        billing_contact_name (str, optional): Billing address
            contact name.
        billing_business_name (str, optional): Billing address
            business name.
        billing_street_1 (str, optional): Billing street address
            (line 1).
        billing_street_2 (str, optional): Billing street address
            (line 2).
        billing_city (str, optional): Billing city.
        billing_province (str, optional): Billing province.
        billing_country (str, optional): Billing country.
        billing_postal_code (str, optional): Billing postal code.
        billing_phone (str, optional): Billing phone number.
        billing_fax (str, optional): Billing fax number
        billing_email (str, optional): Billing email.
        shipping_contact_name (str, optional): Shipping contact
                name.
        shipping_business_name (str, optional): Shipping business
            name.
        shipping_street_1 (str, optional): Shipping street address
            (line 1).
        shipping_street_2 (str, optional): Shipping street address
            (line 2).
        shipping_city (str, optional): Shipping city.
        shipping_province (str, optional): Shipping province.
        shipping_country (str, optional): Shipping country.
        shipping_postal_code (str, optional): Shipping postal code.
        shipping_phone (str, optional): Shipping phone number.
        shipping_fax (str, optional): Shipping fax number.
        shipping_email (str, optional): Shipping email address.
        amount_shipping (dec, optional): Shipping cost.
        amount_tax (dec, optional): Tax amount.
        shipping_method (str, optional): Method of shipping.
        tax_details (str, optional): Name for the tax (e.g. GST).
        test (bool, optional): Whether this is a test transaction or not.
    """

    def __init__(self, api_details=None, **kwargs):
        self.api = self.set_api_details(api_details)
        self.details = kwargs
        self.cleaned = {}
        self.response = {}
        self.redacted_response = {}

    def _redact_api_data(self):
        """Redacts API data and updates redacted_response attribute."""
        if 'raw_request' in self.redacted_response:
            self.redacted_response['raw_request'] = re.sub(
                r'(accountId=.*?)(&|$)',
                r'accountId=REDACTED\g<2>',
                self.redacted_response['raw_request']
            )
            self.redacted_response['raw_request'] = re.sub(
                r'(apiToken=.*?)(&|$)',
                r'apiToken=REDACTED\g<2>',
                self.redacted_response['raw_request']
            )
            self.redacted_response['raw_request'] = re.sub(
                r'(terminalId=.*?)(&|$)',
                r'terminalId=REDACTED\g<2>',
                self.redacted_response['raw_request']
            )
        else:
            self.redacted_response['raw_request'] = None

    def _redact_field(self, api_name, python_name):
        """Redacts all information for the provided field.

            Method directly updates the redacted_response attribute.

            Parameters:
                api_name (str): The field name used by the Helcim API.
                python_name (str): The field name used by this
                    application.
        """
        # Redacts the raw_request data (if present)
        if self.redacted_response.get('raw_request', None):
            self.redacted_response['raw_request'] = re.sub(
                r'({}=.*?)(&|$)'.format(api_name),
                r'{}=REDACTED\g<2>'.format(api_name),
                self.redacted_response['raw_request']
            )

        # Redacts the raw_response data (if present)
        if self.redacted_response.get('raw_response', None):
            self.redacted_response['raw_response'] = re.sub(
                r'<{}>.*</{}>'.format(api_name, api_name),
                r'<{}>REDACTED</{}>'.format(api_name, api_name),
                self.redacted_response['raw_response']
            )

        if python_name in self.redacted_response:
            self.redacted_response[python_name] = None

    def set_api_details(self, details):
        """Sets the API details for this transaction.

            Will either return a dictionary of the API details from the
            provided details argument, or will look to the Django
            settings file.

            Parameters:
                details (dict): A dictionary of the API details.

            Returns:
                dict: The proper API details from the provided data.

            Raises:
                ImproperlyConfigured: A required API setting is not
                    found.
        """
        api_details = {
            'url': None,
            'account_id': None,
            'token': None,
            'terminal_id': None,
        }

        # Tries and retrieves settings form API details
        if details:
            api_details['url'] = details.get('url')
            api_details['account_id'] = details.get('account_id')
            api_details['token'] = details.get('token')
            api_details['terminal_id'] = details.get('terminal_id')

        # For any missing values, use values/defaults from settings
        if api_details['url'] is None:
            api_details['url'] = SETTINGS['api_url']

        if api_details['account_id'] is None:
            api_details['account_id'] = SETTINGS['account_id']

        if api_details['token'] is None:
            api_details['token'] = SETTINGS['api_token']

        if api_details['terminal_id'] is None:
            api_details['terminal_id'] = SETTINGS['terminal_id']

        return api_details

    def configure_test_transaction(self):
        """Adds test flag to post data if HELCIM_API_TEST is True.

            Method applies to the cleaned data (not the raw POST data).
            If the test flag is declared in both the POST data and
            Django settings file, the POST data takes precedence.
        """

        if 'test' not in self.cleaned and SETTINGS['api_test']:
            self.cleaned['test'] = SETTINGS['api_test']

    def redact_data(self):
        """Removes sensitive and identifiable data.

            By default will redact API fields and populate
            redacted_response attribute. Depending on Django settings,
            may also redact other fields.
        """
        # Copy the response data to the redacted file for updating
        self.redacted_response = copy.deepcopy(self.response)

        # Remove any API content
        self._redact_api_data()

        # Identify and redact any other specified fields
        fields = identify_redact_fields()

        for _, redact_field in fields.items():
            if redact_field['redact']:
                for field in redact_field['fields']:
                    self._redact_field(field['api'], field['python'])

    def process_error_response(self, response_message):
        """Returns error response with proper exception type.

            Parameters:
                response_message (str): The API error response message.

            Raises:
                PaymentError: Raised for any errors returned during a
                    purchase or refund transaction.
                HelcimError: Raised for any other unhandled errors.
        """
        exception_message = 'Helcim API request failed: {}'.format(
            response_message
        )

        if isinstance(self, Purchase):
            raise helcim_exceptions.PaymentError(exception_message)

        if isinstance(self, Preauthorize):
            raise helcim_exceptions.PaymentError(exception_message)

        if isinstance(self, Capture):
            raise helcim_exceptions.PaymentError(exception_message)

        if isinstance(self, Refund):
            raise helcim_exceptions.RefundError(exception_message)

        if isinstance(self, Verification):
            raise helcim_exceptions.VerificationError(exception_message)

        raise helcim_exceptions.HelcimError(exception_message)

    def convert_expiry_to_date(self, expiry):
        """Converts a 4 digit expiry to a datetime object.

            Parameters:
                expiry (str): the four digit representation of the
                    credit card expiry

            Returns:
                obj: the expiry as a datetime object.
        """

        year = int(expiry[2:])
        month = int(expiry[:2])
        day = monthrange(year, month)[1]

        return datetime(year, month, day, tzinfo=pytz.timezone('UTC')).date()

    def create_model_arguments(self, transaction_type):
        """Creates dictionary for use as transaction model arguments.

            Takes the redacted_response data and creates a dictionary
            that can be used as the keyword argumetns for the
            HelcimTransaction model.

            Parameters:
                transaction_type (str): Transaction type for this
                    transaction.

            Returns:
                dict: dictionary of the HelcimTransaction arguments.
        """
        response = self.redacted_response

        # Format the transaction_date
        if all([
                response.get('transaction_date'),
                response.get('transaction_time')
        ]):
            date_response = datetime.combine(
                response.get('transaction_date'),
                response.get('transaction_time')
            )
        else:
            date_response = None

        # Format the credit card expiry date (if present)
        if response.get('cc_expiry', None):
            cc_expiry = self.convert_expiry_to_date(
                response['cc_expiry']
            )
        else:
            cc_expiry = None

        return {
            'raw_request': response.get('raw_request'),
            'raw_response': response.get('raw_response'),
            'transaction_success': response.get('transaction_success'),
            'response_message': response.get('response_message'),
            'notice': response.get('notice'),
            'date_response': date_response,
            'transaction_type': transaction_type,
            'transaction_id': response.get('transaction_id'),
            'amount': response.get('amount'),
            'currency': response.get('currency'),
            'cc_name': response.get('cc_name'),
            'cc_number': response.get('cc_number'),
            'cc_expiry': cc_expiry,
            'cc_type': response.get('cc_type'),
            'token': response.get('token'),
            'token_f4l4': response.get('token_f4l4'),
            'avs_response': response.get('avs_response'),
            'cvv_response': response.get('cvv_response'),
            'approval_code': response.get('approval_code'),
            'order_number': response.get('order_number'),
            'customer_code': response.get('customer_code'),
        }

    def post(self, post_data=None):
        """Makes POST to Helcim API and updates response attribute.

        Parameters:
            post_data (dict): The parameters to pass with the POST request.

        Raises:
            ProcessingError: An error occurred connecting or
                communicating with Helcim API.
        """
        # Make the POST request
        try:
            response = requests.post(
                self.api['url'],
                data=post_data,
            )
        except requests.ConnectionError:
            raise helcim_exceptions.ProcessingError(
                'Unable to connect to Helcim API ({})'.format(self.api['url'])
            )

        # Catch any response errors in status code
        if response.status_code != 200:
            raise helcim_exceptions.ProcessingError(
                'Helcim API request failed with status code {}'.format(
                    response.status_code
                )
            )

        # Create the dictionary ('message' is the XML structure object)
        dict_response = xmltodict.parse(response.text)['message']

        # Catch any issues with the API response
        if dict_response['response'] == '0':
            self.process_error_response(dict_response['responseMessage'])

        # Return the response
        self.response = conversions.process_api_response(
            dict_response,
            post_data,
            response.text
        )

    def save_transaction(self, transaction_type):
        """Saves provided transaction data as Django model instance.

            Parameters:
                transaction_type (str): The type of transaction (``s``
                    for purchase (sale), ``p`` for pre-authorization,
                    ``c`` for capture, and ``r`` for refund).

            Returns:
                obj: A Django model instance of the saved data.

            Raises:
                DjangoError: Issue when attempting to save transaction
                    to database.
        """
        # Redacts data if not already done
        if not self.redacted_response:
            self.redact_data()

        model_dictionary = self.create_model_arguments(transaction_type)

        try:
            saved_model = models.HelcimTransaction.objects.create(
                **model_dictionary
            )
        except IntegrityError as error:
            raise helcim_exceptions.DjangoError(
                'Unable to save transaction record: {}'.format(error)
            )
        except ValueError as error:
            raise helcim_exceptions.DjangoError(
                'Unable to save transaction record: {}'.format(error)
            )

        return saved_model

    def validate_fields(self):
        """Validates Helcim API request fields and coerces values."""
        self.cleaned = conversions.validate_request_fields(self.details)

class BaseCardTransaction(BaseRequest):
    """Base class for transactions involving credit card details."""
    def __init__(self, save_token=False, django_user=None, **kwargs):
        """Extends BaseRequest to include save_token and django_user.

            Parameters:
                save_token (bool): Whether the user has requested this
                    token to be saved or not.
                django_user (obj): The Django model for the requesting
                    user.
        """
        super(BaseCardTransaction, self).__init__(**kwargs)
        self.save_token = self._determine_save_token_status(save_token)
        self.django_user = django_user

    def _determine_save_token_status(self, user_decision):
        """Determines if Helcim card token should be saved.

            Parameters:
                user_decision (bool): Whether the user has requested to
                    save this token or not.

            Returns:
                bool: Whether a token should be saved.
        """
        # Check if vault is enabled in settings
        vault_enabled = SETTINGS['enable_token_vault']

        # If yes, check if user requested save
        if vault_enabled:
            return user_decision

        return False

    def determine_card_details(self):
        """Confirms valid payment details and updates self.cleaned.

        Cycles through the provided details to determine the most
        appropriate payment method. If multiple methods provided, only the
        first match is returned (token > customer code > CC number >
        encrypted magnetic strip > magnetic strip).

        Raises:
            ValueError: No valid payment details provided.
        """

        payment_fields = [
            'token', 'customer_code', 'token_f4l4', 'token_f4l4_skip',
            'cc_number', 'cc_expiry', 'cc_name', 'cc_cvv', 'cc_address',
            'cc_postal_code', 'mag_enc', 'mag_enc_serial_number', 'mag'
        ]

        if 'token' in self.cleaned and 'customer_code' in self.cleaned:
            # F4L4 required or it must be explicitly skipped
            f4l4_skip = self.cleaned.get('token_f4l4_skip', False)

            if f4l4_skip:
                payment_fields.remove('token')
                payment_fields.remove('customer_code')
                payment_fields.remove('token_f4l4_skip')

            if 'token_f4l4' in self.cleaned:
                payment_fields.remove('token')
                payment_fields.remove('customer_code')
                payment_fields.remove('token_f4l4')

        elif 'customer_code' in self.cleaned:
            payment_fields.remove('customer_code')

        elif 'cc_number' in self.cleaned and 'cc_expiry' in self.cleaned:
            payment_fields.remove('cc_number')
            payment_fields.remove('cc_expiry')

            # Add any additional CC details (as provided)
            if 'cc_name' in self.cleaned:
                payment_fields.remove('cc_name')

            if 'cc_cvv' in self.cleaned:
                payment_fields.remove('cc_cvv')

            if 'cc_address' in self.cleaned:
                payment_fields.remove('cc_address')

            if 'cc_postal_code' in self.cleaned:
                payment_fields.remove('cc_postal_code')

        elif (
                'mag_enc' in self.cleaned
                and 'mag_enc_serial_number' in self.cleaned
        ):
            payment_fields.remove('mag_enc')
            payment_fields.remove('mag_enc_serial_number')

        elif 'mag' in self.cleaned:
            payment_fields.remove('mag')

        # Raise error if no fields have been removed
        if len(payment_fields) == 13:
            raise ValueError('No valid payment details provided.')

        # Remove any other fields from self.cleaned
        for field in payment_fields:
            self.cleaned.pop(field, None)

    def save_token_to_vault(self):
        """Saves Helcim card token.

            Returns:
                obj: The HelcimToken model instance, or ``None`` (if
                    model not created).
        """
        token = self.response.get('token', None)
        token_f4l4 = self.response.get('token_f4l4', None)

        # Ensure there is a customer code (can't use token without one)
        try:
            customer_code = self.response['customer_code']
        except KeyError:
            raise helcim_exceptions.ProcessingError(
                'Unable to save token - customer code not provided'
            )

        # Ensures a django user is available if it is the identifier
        if SETTINGS['token_vault_identifier'] != 'helcim':
            if self.django_user is None:
                raise helcim_exceptions.ProcessingError(
                    'Unable to save token - user reference not provided'
                )

        if token and token_f4l4:
            token_instance, _ = models.HelcimToken.objects.get_or_create(
                token=token,
                token_f4l4=token_f4l4,
                cc_type=self.response.get('cc_type', None),
                customer_code=customer_code,
                django_user=self.django_user,
            )

            return token_instance

        return None

class Purchase(BaseCardTransaction):
    """Makes a purchase request to Helcim Commerce API."""
    def process(self):
        """Makes a purchase request.

            Returns:
                purchase (obj): The saved HelcimTransaction model
                    instance.
                token (obj): The saved HelcimToken model.
        """
        self.validate_fields()
        self.configure_test_transaction()
        self.determine_card_details()

        purchase_data = conversions.process_request_fields(
            self.api,
            self.cleaned,
            {
                'transactionType': 'purchase',
            }
        )
        self.post(purchase_data)

        purchase = self.save_transaction('s')

        if self.save_token:
            token = self.save_token_to_vault()
        else:
            token = None

        return purchase, token

class Preauthorize(BaseCardTransaction):
    """Makes a pre-authorization request to Helcim Commerce API."""
    def process(self):
        """Makes a pre-authorization request."""
        self.validate_fields()
        self.configure_test_transaction()
        self.determine_card_details()

        preauth_data = conversions.process_request_fields(
            self.api,
            self.cleaned,
            {
                'transactionType': 'preauth',
            }
        )
        self.post(preauth_data)
        preauth = self.save_transaction('p')

        if self.save_token:
            token = self.save_token_to_vault()
        else:
            token = None

        return preauth, token

class Refund(BaseCardTransaction):
    """Makes a refund request."""
    def process(self):
        """Makes a refund request to Helcim Commerce API."""
        self.validate_fields()
        self.configure_test_transaction()
        self.determine_card_details()

        refund_data = conversions.process_request_fields(
            self.api,
            self.cleaned,
            {
                'transactionType': 'refund',
            }
        )
        self.post(refund_data)
        refund = self.save_transaction('r')

        if self.save_token:
            token = self.save_token_to_vault()
        else:
            token = None


        return refund, token

class Verification(BaseCardTransaction):
    """Makes a verification request to Helcim Commerce API."""
    def process(self):
        """Makes a verification request to Helcim Commerce API."""
        self.validate_fields()
        self.configure_test_transaction()
        self.determine_card_details()

        verification_data = conversions.process_request_fields(
            self.api,
            self.cleaned,
            {
                'transactionType': 'verify',
            }
        )
        self.post(verification_data)
        verification = self.save_transaction('v')

        if self.save_token:
            token = self.save_token_to_vault()
        else:
            token = None


        return verification, token

class Capture(BaseRequest):
    """Makes a capture request (to complete a preauthorization)."""
    def validate_preauth_transaction(self):
        """Confirms that a preauth transaction ID was provided.

            Raises:
                PaymentError: Error when no transaction ID provided.
        """

        if 'transaction_id' not in self.details:
            raise helcim_exceptions.PaymentError(
                'Transaction ID must be provided with capture (force) request.'
            )

    def process(self):
        """Completes a capture request."""
        self.validate_fields()
        self.validate_preauth_transaction()
        self.configure_test_transaction()

        capture_data = conversions.process_request_fields(
            self.api,
            self.cleaned,
            {
                'transactionType': 'capture',
            }
        )

        self.post(capture_data)
        capture = self.save_transaction('c')

        return capture

def identify_redact_fields():
    """Identifies which fields (if any) should be redacted.

        Configured using flags in the Django settings file.

        Returns:
            dict: A dictionary of fields to redact with corresponding
                API field and variable names.
    """
    redact_fields = {
        'name': {
            'redact': None,
            'settings': 'redact_cc_name',
            'fields': [
                {'api': 'cardHolderName', 'python': 'cc_name'},
            ]
        },
        'number': {
            'redact': None,
            'settings': 'redact_cc_number',
            'fields': [
                {'api': 'cardNumber', 'python': 'cc_number'},
            ]
        },
        'expiry': {
            'redact': None,
            'settings': 'redact_cc_expiry',
            'fields': [
                {'api': 'cardExpiry', 'python': 'cc_expiry'},
                {'api': 'expiryDate', 'python': 'cc_expiry'},
            ]
        },
        'cvv': {
            'redact': None,
            'settings': 'redact_cc_cvv',
            'fields': [
                {'api': 'cardCVV', 'python': 'cc_cvv'},
            ]
        },
        'type': {
            'redact': None,
            'settings': 'redact_cc_type',
            'fields': [
                {'api': 'cardType', 'python': 'cc_type'},
            ]
        },
        'token': {
            'redact': None,
            'settings': 'redact_token',
            'fields': [
                {'api': 'cardToken', 'python': 'token'},
                {'api': 'cardF4L4', 'python': 'token_f4l4'},
            ]
        },
        'mag': {
            'redact': None,
            'settings': 'redact_cc_magnetic',
            'fields': [
                {'api': 'cardMag', 'python': 'mag'},
            ]
        },
        'mag_enc': {
            'redact': None,
            'settings': 'redact_cc_magnetic_encrypted',
            'fields': [
                {'api': 'cardMagEnc', 'python': 'mag_enc'},
                {'api': 'serialNumber', 'python': 'mang_enc_serial_number'},
            ]
        },
    }

    # HELCIM_REDACT_ALL overrides all other settings
    if SETTINGS['redact_all'] is not None:
        if SETTINGS['redact_all'] is True:
            for key in redact_fields:
                redact_fields[key]['redact'] = True
        else:
            for key in redact_fields:
                redact_fields[key]['redact'] = False

    # Otherwise, assess each field individually
    else:
        for key, field in redact_fields.items():
            redact_fields[key]['redact'] = SETTINGS[field['settings']]

    return redact_fields

def determine_helcim_settings():
    """Collects all possible django-helcim settings for easy use.

        Performs basic validation of required settings and assigns
        defaults where applicable.

        Returns:
            dict: Summary of all possible django-helcim settings.
    """
    # API SETTINGS
    # -------------------------------------------------------------------------
    # Required settings
    try:
        account_id = getattr(django_settings, 'HELCIM_ACCOUNT_ID')
    except AttributeError:
        raise django_exceptions.ImproperlyConfigured(
            'You must define a HELCIM_ACCOUNT_ID setting'
        )

    try:
        api_token = getattr(django_settings, 'HELCIM_API_TOKEN')
    except AttributeError:
        raise django_exceptions.ImproperlyConfigured(
            'You must define a HELCIM_API_TOKEN setting'
        )

    # Other settings
    api_url = getattr(
        django_settings, 'HELCIM_API_URL', 'https://secure.myhelcim.com/api/'
    )
    terminal_id = getattr(django_settings, 'HELCIM_TERMINAL_ID', '')
    api_test = getattr(django_settings, 'HELCIM_API_TEST', None)

    # REDACTION SETTINGS
    # -------------------------------------------------------------------------
    redact_all = getattr(django_settings, 'HELCIM_REDACT_ALL', None)
    redact_cc_name = getattr(django_settings, 'HELCIM_REDACT_CC_NAME', True)
    redact_cc_number = getattr(
        django_settings, 'HELCIM_REDACT_CC_NUMBER', True
    )
    redact_cc_expiry = getattr(
        django_settings, 'HELCIM_REDACT_CC_EXPIRY', True
    )
    redact_cc_cvv = getattr(django_settings, 'HELCIM_REDACT_CC_CVV', True)
    redact_cc_type = getattr(django_settings, 'HELCIM_REDACT_CC_TYPE', True)
    redact_cc_magnetic = getattr(
        django_settings, 'HELCIM_REDACT_CC_MAGNETIC', True
    )
    redact_cc_magnetic_encrypted = getattr(
        django_settings, 'HELCIM_REDACT_CC_MAGNETIC_ENCRYPTED', True
    )
    redact_token = getattr(django_settings, 'HELCIM_REDACT_TOKEN', False)

    # TRANSACTION FUNCTIONALITY SETTINGS
    # -------------------------------------------------------------------------
    enable_transaction_capture = getattr(
        django_settings, 'HELCIM_ENABLE_TRANSACTION_CAPTURE', False
    )
    enable_transaction_refund = getattr(
        django_settings, 'HELCIM_ENABLE_TRANSACTION_REFUND', False
    )

    # TOKEN VAULT SETTINGS
    # -------------------------------------------------------------------------
    enable_token_vault = getattr(
        django_settings, 'HELCIM_ENABLE_TOKEN_VAULT', False
    )
    token_vault_identifier = getattr(
        django_settings, 'HELCIM_TOKEN_VAULT_IDENTIFIER', 'django'
    )

    # ADMIN SETTINGS
    # -------------------------------------------------------------------------
    enable_admin = getattr(
        django_settings, 'HELCIM_ENABLE_ADMIN', False
    )

    return {
        'api_url': api_url,
        'account_id': account_id,
        'api_token': api_token,
        'terminal_id': terminal_id,
        'api_test': api_test,
        'redact_all': redact_all,
        'redact_cc_name': redact_cc_name,
        'redact_cc_number': redact_cc_number,
        'redact_cc_expiry': redact_cc_expiry,
        'redact_cc_cvv': redact_cc_cvv,
        'redact_cc_type': redact_cc_type,
        'redact_cc_magnetic': redact_cc_magnetic,
        'redact_cc_magnetic_encrypted': redact_cc_magnetic_encrypted,
        'redact_token': redact_token,
        'enable_transaction_capture': enable_transaction_capture,
        'enable_transaction_refund': enable_transaction_refund,
        'enable_token_vault': enable_token_vault,
        'token_vault_identifier': token_vault_identifier,
        'enable_admin': enable_admin,
    }

def retrieve_token_details(token_id, customer):
    """Takes a HelcimToken ID and maps details to dictionary."""
    # Final validation to ensure token exists & belongs to proper user
    try:
        # Determine what identifier is used for tokens
        customer_reference = SETTINGS['token_vault_identifier']

        if customer_reference == 'helcim':
            token_instance = models.HelcimToken.objects.get(
                id=token_id, customer_code=customer
            )
        else:
            token_instance = models.HelcimToken.objects.get(
                id=token_id, django_user=customer
            )
    except models.HelcimToken.DoesNotExist:
        raise helcim_exceptions.ProcessingError(
            'Unable to retrieve token details for specified customer.'
        )

    # Validation passed - map model to the dictionary
    return {
        'token': token_instance.token,
        'token_f4l4': token_instance.token_f4l4,
        'customer_code': token_instance.customer_code,
    }

def retrieve_saved_tokens(customer):
    """Returns list of tokens for specified customer.

        Parameters:
            customer: Either a a Django user instance or a Helcim
                customer code. The reference used is determined by the
                ``HELCIM_TOKEN_VAULT_IDENTIFIER`` setting.

        Returns:
            obj: A queryset of the retrieved tokens
    """
    if SETTINGS['token_vault_identifier'] == 'helcim':
        tokens = models.HelcimToken.objects.filter(customer_code=customer)
    else:
        tokens = models.HelcimToken.objects.filter(django_user=customer)

    return tokens

SETTINGS = determine_helcim_settings()



"""Bridging module between Django Oscar and gateway module."""
import logging

from oscar.apps.payment import exceptions as oscar_exceptions

from helcim import exceptions as helcim_exceptions, gateway

LOG = logging.getLogger(__name__)

def remap_oscar_billing_address(address):
    """Remaps Oscar billing address dictionary to Helcim dictionary.

        Parameters:
            address (dict): A dictionary of the billing address details
                provided by Django Oscar.

        Returns:
            dict: Billing address details formatted for
                django-helcim.
    """
    if address.get('first_name') and address.get('last_name'):
        contact_name = '{} {}'.format(
            address.get('first_name'), address.get('last_name')
        )
    elif address.get('first_name'):
        contact_name = address.get('first_name')
    elif address.get('last_name'):
        contact_name = address.get('last_name')
    else:
        contact_name = None

    return {
        'billing_contact_name': contact_name,
        'billing_street_1': address.get('line1'),
        'billing_street_2': address.get('line2'),
        'billing_city': address.get('line4'),
        'billing_province': address.get('state'),
        'billing_postal_code': address.get('postcode'),
        'billing_country': address.get('country'),
        'billing_phone': address.get('phone_number'),
    }

def remap_oscar_credit_card(card):
    """Remaps Oscar credit card object as Helcim dictionary.

        Parameters:
            card (obj): A credit card object provided by Django Oscar.

        Returns:
            dict: Credit card details formatted for django-helcim.
    """
    if card.expiry_date:
        cc_expiry = card.expiry_date.strftime('%m%y')
    else:
        cc_expiry = None

    return {
        'cc_name': card.name,
        'cc_number': card.number,
        'cc_expiry': cc_expiry,
        'cc_cvv': card.ccv,
    }

class BaseCardTransactionBridge():
    """Base class to bridge Oscar and Helcim transactions.

        Parameters:
            order_number (str): Order number for the transaction.
            amount (dec): The transaction total.
            card (obj): Instance of the Oscar bankcard class.
            billing_address (dict): The billing address information.
            save_token (bool): Whether the card token should be saved
                in the token vault or not.
            django_user (obj): The user to associate with the saved
                card token.
            customer_code (str): The Helcim customer code to associate
                with the saved card token.
    """
    def __init__(
            self, amount, token_id=None, card=None,
            billing_address=None, save_token=False, django_user=None,
            customer_code=None
    ):
        transaction_details = {
            'amount': amount,
        }

        if token_id:
            if gateway.SETTINGS['token_vault_identifier'] == 'helcim':
                transaction_details.update(retrieve_token_details(
                    token_id, customer_code
                ))
            else:
                transaction_details.update(retrieve_token_details(
                    token_id, django_user
                ))

        if billing_address:
            transaction_details.update(remap_oscar_billing_address(
                billing_address
            ))

        if card:
            transaction_details.update(remap_oscar_credit_card(card))

        self.transaction_details = transaction_details
        self.save_token = save_token
        self.django_user = django_user

class PurchaseBridge(BaseCardTransactionBridge):
    """Class to bridge Oscar and Helcim purchase transactions."""
    def process(self):
        """Attempts to process Purchase with transaction details.

            Returns the values of ``gateway.Purchase.process``.

            Raises:
                GatewayError: An Oscar error raised when there was an
                    error with the payment API.
                PaymentError: An Oscar error raised when there was an
                    error processing the payment.
        """
        purchase_instance = gateway.Purchase(
            save_token=self.save_token,
            django_user=self.django_user,
            **self.transaction_details
        )

        try:
            return purchase_instance.process()
        except helcim_exceptions.ProcessingError as error:
            raise oscar_exceptions.GatewayError(str(error))
        except helcim_exceptions.PaymentError as error:
            raise oscar_exceptions.PaymentError(str(error))
        except helcim_exceptions.DjangoError:
            LOG.exception(
                'Purchase complete, but errors occured while saving '
                'transaction to database'
            )
        except helcim_exceptions.HelcimError as error:
            raise oscar_exceptions.GatewayError(str(error))

class PreauthorizeBridge(BaseCardTransactionBridge):
    """Class to bridge Oscar and Helcim preauthorization transactions."""
    def process(self):
        """Attempts to process Preauthorize with transaction details.

            Returns the values of ``gateway.Preauthorize.process``.

            Raises:
                GatewayError: An Oscar error raised when there was an
                    error with the payment API.
                PaymentError: An Oscar error raised when there was an
                    error processing the payment.
        """

        preauth_instance = gateway.Preauthorize(
            save_token=self.save_token,
            django_user=self.django_user,
            **self.transaction_details
        )

        try:
            return preauth_instance.process()
        except helcim_exceptions.ProcessingError as error:
            raise oscar_exceptions.GatewayError(str(error))
        except helcim_exceptions.PaymentError as error:
            raise oscar_exceptions.PaymentError(str(error))
        except helcim_exceptions.DjangoError:
            LOG.exception(
                'Preauthorization complete, but errors occured while saving '
                'transaction to database'
            )
        except helcim_exceptions.HelcimError as error:
            raise oscar_exceptions.GatewayError(str(error))

class RefundBridge(BaseCardTransactionBridge):
    """Class to bridge Oscar and Helcim refund transactions."""
    def process(self):
        """Attempts to process Refund with transaction details.

            Returns the values of ``gateway.Refund.process``.

            Raises:
                GatewayError: An Oscar error raised when there was an
                    error with the payment API.
                PaymentError: An Oscar error raised when there was an
                    error processing the payment.
        """

        refund_instance = gateway.Refund(
            save_token=self.save_token,
            django_user=self.django_user,
            **self.transaction_details
        )

        try:
            return refund_instance.process()
        except helcim_exceptions.ProcessingError as error:
            raise oscar_exceptions.GatewayError(str(error))
        except helcim_exceptions.PaymentError as error:
            raise oscar_exceptions.PaymentError(str(error))
        except helcim_exceptions.DjangoError:
            LOG.exception(
                'Refund complete, but errors occured while saving '
                'transaction to database'
            )
        except helcim_exceptions.HelcimError as error:
            raise oscar_exceptions.GatewayError(str(error))

class VerificationBridge(BaseCardTransactionBridge):
    """Class to bridge Oscar and Helcim Verification transactions."""
    def process(self):
        """Attempts to process Verification with transaction details.

            Returns the values of ``gateway.Verification.process``.

            Raises:
                GatewayError: An Oscar error raised when there was an
                    error with the payment API.
                PaymentError: An Oscar error raised when there was an
                    error processing the payment.
        """

        verification_instance = gateway.Verification(
            save_token=self.save_token,
            django_user=self.django_user,
            **self.transaction_details
        )

        try:
            return verification_instance.process()
        except helcim_exceptions.ProcessingError as error:
            raise oscar_exceptions.GatewayError(str(error))
        except helcim_exceptions.PaymentError as error:
            raise oscar_exceptions.PaymentError(str(error))
        except helcim_exceptions.DjangoError:
            LOG.exception(
                'Verification complete, but errors occured while saving '
                'transaction to database'
            )
        except helcim_exceptions.HelcimError as error:
            raise oscar_exceptions.GatewayError(str(error))

class CaptureBridge():
    """Class to bridge Oscar and Helcim Capture transactions.

        Parameters:
            transaction_id (int): the Helcim preauthorization
                 transaction ID to capture.
    """
    def __init__(self, transaction_id):
        transaction_details = {
            'transaction_id': transaction_id,
        }

        self.transaction_details = transaction_details

    def process(self):
        """Attempts to process Capture with transaction details.

            Returns the values of ``gateway.Capture.process``.

            Raises:
                GatewayError: An Oscar error raised when there was an
                    error with the payment API.
                PaymentError: An Oscar error raised when there was an
                    error processing the payment.
        """

        capture_instance = gateway.Capture(
            **self.transaction_details
        )

        try:
            return capture_instance.process()
        except helcim_exceptions.ProcessingError as error:
            raise oscar_exceptions.GatewayError(str(error))
        except helcim_exceptions.PaymentError as error:
            raise oscar_exceptions.PaymentError(str(error))
        except helcim_exceptions.DjangoError:
            LOG.exception(
                'Capture complete, but errors occured while saving '
                'transaction to database'
            )
            return None
        except helcim_exceptions.HelcimError as error:
            raise oscar_exceptions.GatewayError(str(error))


# SHORTCUT FUNCTIONS AND DICTIONARIES
# -----------------------------------------------------------------------------
# These functions/dictionaries are provided as a convenience shortcut to allow
# access to main functions via the bridge module.

def retrieve_token_details(token_id, customer):
    """Shortcut for retrieve_token_details from the Gateway module.

        Added as a convenience to allow access to core functions via
        the bridge module exclusively.
    """
    return gateway.retrieve_token_details(token_id, customer)

def retrieve_saved_tokens(customer):
    """Shortcut for retrieve_saved_tokens from the Gateway module.

        Added as a convenience to allow access to core functions via
        the bridge module exclusively.
    """
    return gateway.retrieve_saved_tokens(customer)

SETTINGS = gateway.SETTINGS
